#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <defines.h>
#include <audiobuffer.h>
#include <dspoutoss.h>

DspOutOss::DspOutOss( const QString &fileName )
	: DspOut()
{
	filename = fileName;
}

DspOutOss::~DspOutOss()
{
}

bool DspOutOss::openDevice()
{
	if( devstate == DeviceOpened ) {
		lasterror = "Device Already Open";
		return false;
	}

	// Open immediately or else
	audio_fd = open( filename.ascii(), O_WRONLY | O_NONBLOCK );
	if( audio_fd == -1 ) {
		lasterror = "Open Failed";
		return false;
	}

	// Remove O_NONBLOCK
	int flags = fcntl( audio_fd, F_GETFL );
	flags &= ~O_NONBLOCK;
	fcntl( audio_fd, F_SETFL, flags );

	int format = AFMT_S16_LE;

	if( ioctl( audio_fd, SNDCTL_DSP_SETFMT, &format ) == -1 ) {
		lasterror = sys_errlist[ errno ];
		return false;
	}

	if( format != AFMT_S16_LE ) {
		lasterror = "Format not supported";
		return false;
	}

	if( ioctl( audio_fd, SNDCTL_DSP_SPEED, &rate ) == -1 ) {
		lasterror = sys_errlist[ errno ];
		return false;
	}

	int chanreturn = channels - 1;

	if( ioctl( audio_fd, SNDCTL_DSP_STEREO, &chanreturn ) == -1 ) {
		lasterror = sys_errlist[ errno ];
		return false;
	}

	if( ( channels - 1 ) != chanreturn ) {
		lasterror = "Unsupported Number of Channels";
		return false;
	}

	int numfrags = 12;
	int size = 10;

	audio_buf_info info;
	int frag = ( ( ( numfrags - 1 ) << 16 ) | size );

	if( ioctl( audio_fd, SNDCTL_DSP_SETFRAGMENT, &frag ) ) {
		lasterror = sys_errlist[ errno ];
		return false;
	}

	if( ioctl( audio_fd, SNDCTL_DSP_GETOSPACE, &info ) == -1 ) {
		lasterror = sys_errlist[ errno ];
		return false;
	}

	totalfrags = info.fragments;
	audio_buf = new AudioBuffer( info.fragsize * sizeof( sample_t ) );
	output_buf = new AudioBuffer( info.fragsize * sizeof( unsigned short ) );

	lasterror = QString::null;
	devstate = DeviceOpened;
	return true;
}

bool DspOutOss::closeDevice()
{
	if( devstate != DeviceOpened ) {
		lasterror = "Device Not Open";
		return false;
	}
	close( audio_fd );
	return true;
}

unsigned int DspOutOss::writableFragments()
{
	struct timeval timeout;
	audio_buf_info info;
	fd_set write_fds;

	if( devstate != DeviceOpened ) {
		return 0;
	}

	timeout.tv_sec = 0;
	timeout.tv_usec = 0;

	FD_ZERO( &write_fds );
	FD_SET( audio_fd, &write_fds );

	if( select( audio_fd + 1, NULL, &write_fds, NULL, &timeout ) == -1 ) {
		return 0;
	}

	if( !FD_ISSET( audio_fd, &write_fds ) ) {
		return 0;
	}

	if( ioctl( audio_fd, SNDCTL_DSP_GETOSPACE, &info ) == -1 ) {
		lasterror = sys_errlist[ errno ];
		return 0;
	}

	return info.fragments;
}

bool DspOutOss::writeBuffer()
{
	if( devstate != DeviceOpened ) {
		lasterror = "Device Not Open";
		return false;
	}

	sample_t *inbuf;
	unsigned short *outbuf;
	sample_t iterate;
	unsigned int i;

	inbuf = (sample_t *) audio_buf->getData();
	outbuf = (unsigned short *) output_buf->getData();

	for( i = 0; i < ( audio_buf->getSize() / sizeof( sample_t ) ); i++ ) {
		iterate = *inbuf;
		inbuf++;

		if( iterate > 32767.0 ) { iterate = 32767.0; }
		if( iterate < -32767.0 ) { iterate = -32767.0; }

		*outbuf = (unsigned short) iterate;
		outbuf++;
	}

	for(;;) {
		if( write( audio_fd, output_buf->getData(), output_buf->getSize() ) != -1 ) {
			break;
		}
	}

	return true;
}

bool DspOutOss::setNumChannels( int newchans )
{
	if( devstate == DeviceOpened ) {
		lasterror =  "Can't reset number of channels while device open";
		return false;
	}

	channels = newchans;
	return true;
}

bool DspOutOss::setSampleRate( int newrate )
{
	if( devstate == DeviceOpened ) {
		lasterror = "Can't reset rate while device open";
		return false;
	}

	rate = newrate;
	return true;
}

